@@ -67,6 +67,10 @@ class AppmenusSubdirs:
 class Appmenus(object):
     """Main class for menu entries handling"""
 
+    qubes_vm_desktop = 'org.qubes-os.vm'
+    qubes_dispvm_desktop = 'org.qubes-os.dispvm'
+    qubes_vm_desktop_settings = 'org.qubes-os.qubes-vm-settings'
+
     def templates_dirs(self, vm, template=None):
         """
 
@@ -121,6 +125,8 @@ class Appmenus(object):
         """File name of desktop directory entry template"""
         if dispvm:
             return 'qubes-dispvm.directory.template'
+        if getattr(vm, 'template_for_dispvms', False):
+            return 'qubes-templatedispvm.directory.template'
         if vm.klass == 'TemplateVM':
             return 'qubes-templatevm.directory.template'
         if vm.provides_network:
@@ -146,7 +152,11 @@ class Appmenus(object):
             source = source. \
                 replace('\nExec=', '\nX-Qubes-NonDispvmExec='). \
                 replace('\nX-Qubes-DispvmExec=', '\nExec=')
-        icon = vm.icon
+        if dispvm:
+            # menu directory for creating new DispVMs is special
+            icon = 'dispvm-' + vm.label.name
+        else:
+            icon = vm.icon
         data = source. \
             replace("%VMNAME%", vm.name). \
             replace("%VMDIR%", os.path.join(basedir, vm.name)). \
@@ -210,6 +220,70 @@ class Appmenus(object):
                     result.append(field_values.get(field, ''))
             yield result
 
+    def desktop_name(self, vm, appmenu_basename: str, dispvm=False):
+        """Return the basename of a ``.desktop`` file.
+
+        :param vm: QubesVM object for the VM whose entry this is.
+        :param appmenu_basename: The basename of the appmenu, as provided by
+        the VM.  This is not marked as “untrusted” as it has already been
+        validated.
+        """
+        assert appmenu_basename.endswith('.desktop')
+        if dispvm:
+            prefix = self.qubes_dispvm_desktop
+        else:
+            prefix = self.qubes_vm_desktop
+        return '.'.join((prefix, str(vm), appmenu_basename))
+
+    def settings_name(self, vm):
+        """Return the basename of the .desktop file for the “Qube Settings”
+        menu entry of a given VM..
+
+        :param vm: QubesVM object for the VM.
+        """
+        return '.'.join((self.qubes_vm_desktop_settings, str(vm), 'desktop'))
+
+    def _do_remove_appmenus(self, vm, appmenus_to_remove, appmenus_dir,
+            refresh_cache):
+        """Remove old appmenus
+
+        :param vm: QubesVM object for the VM.
+        :param appmenus_to_remove: List of appmenus to remove.
+        :param appmenus_dir: Appmenus directory.
+        :param refresh_cache: Refresh the cache?
+        """
+        old_appmenus = set((fname for fname in appmenus_to_remove
+                                  if self._is_old_path(fname)))
+        new_appmenus = set(appmenus_to_remove).difference(old_appmenus)
+
+        def _sort_entries(entry):
+            return sorted(entry, key=lambda x: not x.endswith('.directory'))
+
+        for bad_menus in (_sort_entries(old_appmenus),
+                          _sort_entries(new_appmenus)):
+            if bad_menus:
+                appmenus_to_remove_fnames = [os.path.join(appmenus_dir, x)
+                                             for x in bad_menus]
+                try:
+                    desktop_menu_cmd = ['xdg-desktop-menu', 'uninstall']
+                    if not refresh_cache:
+                        desktop_menu_cmd.append('--noupdate')
+                    desktop_menu_cmd.extend(appmenus_to_remove_fnames)
+                    desktop_menu_env = os.environ.copy()
+                    desktop_menu_env['LC_COLLATE'] = 'C'
+                    subprocess.check_call(desktop_menu_cmd,
+                                          env=desktop_menu_env)
+                except subprocess.CalledProcessError:
+                    if hasattr(vm, 'log'):
+                        vm.log.warning(
+                            "Problem removing appmenus for %s", vm.name)
+
+                    else:
+                        print(
+                            "Problem removing appmenus for {!s}".format(vm),
+                            file=sys.stderr)
+
+
     def appmenus_create(self, vm, force=False, refresh_cache=True):
         """Create/update .desktop files
 
@@ -227,18 +301,42 @@ class Appmenus(object):
 
         if hasattr(vm, 'log'):
             vm.log.info("Creating appmenus")
+
+        try:
+            dispvm = (vm.template_for_dispvms and
+                      vm.features.get('appmenus-dispvm', False))
+        except (qubesadmin.exc.QubesDaemonNoResponseError,
+                qubesadmin.exc.QubesNoSuchPropertyError):
+            dispvm = False
+        self._appmenus_create_onedir(
+            vm, force=force, refresh_cache=refresh_cache, dispvm=False,
+            keep_dispvm=dispvm)
+
+        if dispvm:
+            self._appmenus_create_onedir(
+                vm, force=force, refresh_cache=refresh_cache, dispvm=True)
+
+    def _appmenus_create_onedir(self, vm, force=False, refresh_cache=True,
+                                dispvm=False, keep_dispvm=False):
+        """Create/update .desktop files
+
+        :param vm: QubesVM object for which create entries
+        :param refresh_cache: refresh desktop environment cache; if false,
+        must be refreshed manually later
+        :param force: force re-registering files even if unchanged
+        :param dispvm: whether create entries for launching new DispVM
+        :param keep_dispvm: do not remove DispVM-related files,
+        to be used together with dispvm=False
+        :return: None
+        """
+
         appmenus_dir = self.appmenus_dir(vm)
         if not os.path.exists(appmenus_dir):
             os.makedirs(appmenus_dir)
 
-        try:
-            dispvm = vm.features.get('appmenus-dispvm', False)
-        except qubesadmin.exc.QubesDaemonNoResponseError:
-            dispvm = False
-
         anything_changed = False
         directory_changed = False
-        directory_file = os.path.join(appmenus_dir, vm.name + '-vm.directory')
+        directory_file = self._directory_path(vm, dispvm=dispvm)
         if self.write_desktop_file(
                 vm,
                 pkg_resources.resource_string(
@@ -260,62 +358,68 @@ class Appmenus(object):
 
         for appmenu in appmenus:
             appmenu_basename = os.path.basename(appmenu)
+            fname = os.path.join(appmenus_dir,
+                                 self.desktop_name(vm, appmenu_basename,
+                                                   dispvm=dispvm))
             try:
-                if self.write_desktop_file(
-                        vm,
-                        appmenu,
-                        os.path.join(appmenus_dir,
-                                     '-'.join((vm.name, appmenu_basename))),
-                        dispvm):
-                    changed_appmenus.append(appmenu_basename)
+                if self.write_desktop_file(vm, appmenu, fname, dispvm):
+                    changed_appmenus.append(fname)
             except DispvmNotSupportedError:
                 # remove DispVM-incompatible entries
                 with contextlib.suppress(FileNotFoundError):
-                    os.unlink(os.path.join(
-                        appmenus_dir, '-'.join((vm.name, appmenu_basename))))
-        if self.write_desktop_file(
-                vm,
-                pkg_resources.resource_string(
-                    __name__,
-                    'qubes-vm-settings.desktop.template'
-                ).decode(),
-                os.path.join(appmenus_dir,
-                             '-'.join((vm.name, 'qubes-vm-settings.desktop')))):
-            changed_appmenus.append('qubes-vm-settings.desktop')
+                    os.unlink(fname)
+
+        target_appmenus = [
+            self.desktop_name(vm, os.path.basename(x), dispvm=dispvm)
+            for x in appmenus]
+
+        if not dispvm:
+            vm_settings_fname = os.path.join(
+                appmenus_dir, self.settings_name(vm))
+            if self.write_desktop_file(
+                    vm,
+                    pkg_resources.resource_string(
+                        __name__,
+                        'qubes-vm-settings.desktop.template'
+                    ).decode(),
+                    vm_settings_fname):
+                changed_appmenus.append(vm_settings_fname)
+            target_appmenus.append(os.path.basename(vm_settings_fname))
 
         if changed_appmenus:
             anything_changed = True
 
-        target_appmenus = ['-'.join((vm.name, os.path.basename(x)))
-                           for x in appmenus + ['qubes-vm-settings.desktop']]
-
         # remove old entries
         installed_appmenus = os.listdir(appmenus_dir)
         installed_appmenus.remove(os.path.basename(directory_file))
         appmenus_to_remove = set(installed_appmenus).difference(set(
             target_appmenus))
-        if appmenus_to_remove:
-            appmenus_to_remove_fnames = [os.path.join(appmenus_dir, x)
-                                         for x in appmenus_to_remove]
-            try:
-                desktop_menu_cmd = ['xdg-desktop-menu', 'uninstall']
-                if not refresh_cache:
-                    desktop_menu_cmd.append('--noupdate')
-                desktop_menu_cmd.extend(appmenus_to_remove_fnames)
-                desktop_menu_env = os.environ.copy()
-                desktop_menu_env['LC_COLLATE'] = 'C'
-                subprocess.check_call(desktop_menu_cmd, env=desktop_menu_env)
-            except subprocess.CalledProcessError:
-                vm.log.warning("Problem removing old appmenus")
+        if keep_dispvm:
+            appmenus_to_remove = [
+                x for x in appmenus_to_remove
+                if not x.startswith(self.qubes_dispvm_desktop) and
+                   not x.startswith('qubes-dispvm-directory-')]
+        elif dispvm:
+            appmenus_to_remove = [
+                x for x in appmenus_to_remove
+                if not x.startswith(self.qubes_vm_desktop) and
+                   not x.startswith('qubes-vm-directory-')]
+        self._do_remove_appmenus(vm, appmenus_to_remove, appmenus_dir,
+                refresh_cache)
 
-            for appmenu in appmenus_to_remove_fnames:
-                try:
-                    os.unlink(appmenu)
-                except FileNotFoundError:
-                    pass
+        for appmenu in (os.path.join(appmenus_dir, x)
+                for x in appmenus_to_remove):
+            try:
+                if hasattr(vm, 'log'):
+                    vm.log.warning('Removing appmenu %r', appmenu)
+                os.unlink(appmenu)
+            except FileNotFoundError:
+                pass
 
         # add new entries
         if anything_changed or force:
+            # Only call ‘xdg-desktop-menu’ if it has at least one file argument
+            do_anything = False
             try:
                 desktop_menu_cmd = ['xdg-desktop-menu', 'install']
                 if not refresh_cache:
@@ -324,16 +428,20 @@ class Appmenus(object):
                 if (directory_changed and not changed_appmenus) or force:
                     # only directory file changed, not actual entries;
                     # re-register all of them to force refresh
-                    desktop_menu_cmd.extend(
-                        os.path.join(appmenus_dir, x)
-                        for x in target_appmenus)
-                else:
-                    desktop_menu_cmd.extend(
-                        os.path.join(appmenus_dir, '-'.join((vm.name, x)))
-                        for x in changed_appmenus)
-                desktop_menu_env = os.environ.copy()
-                desktop_menu_env['LC_COLLATE'] = 'C'
-                subprocess.check_call(desktop_menu_cmd, env=desktop_menu_env)
+                    if target_appmenus:
+                        desktop_menu_cmd.extend(
+                            os.path.join(appmenus_dir, x)
+                            for x in target_appmenus)
+                        do_anything = True
+                elif changed_appmenus:
+                    desktop_menu_cmd.extend(changed_appmenus)
+                    do_anything = True
+                if do_anything:
+                    desktop_menu_env = os.environ.copy()
+                    desktop_menu_env['LC_COLLATE'] = 'C'
+                    subprocess.check_call(
+                        desktop_menu_cmd,
+                        env=desktop_menu_env)
             except subprocess.CalledProcessError:
                 vm.log.warning("Problem creating appmenus for %s", vm.name)
 
@@ -342,6 +450,28 @@ class Appmenus(object):
                 subprocess.call(['kbuildsycoca' +
                                  os.environ.get('KDE_SESSION_VERSION', '4')])
 
+    @staticmethod
+    def _is_old_path(name):
+        """Return if a path to a desktop name is for the old menu version.
+        """
+        return not (name.startswith('org.qubes-os.') or
+                    name.startswith('qubes-vm-directory-'))
+
+    def _old_directory_path(self, vm):
+        """Return the old path of the directory file for this VM
+        """
+        return os.path.join(self.appmenus_dir(vm), str(vm) + '.vm.directory')
+
+    def _directory_path(self, vm, dispvm=False):
+        """Return the path of the directory file for this VM
+        """
+        if dispvm:
+            basename = (
+               'qubes-dispvm-directory-' + str(vm) + '.directory')
+        else:
+            basename = 'qubes-vm-directory-' + str(vm) + '.directory'
+        return os.path.join(self.appmenus_dir(vm), basename)
+
     def appmenus_remove(self, vm, refresh_cache=True):
         """Remove desktop files for particular VM
 
@@ -357,31 +487,8 @@ class Appmenus(object):
                     print("Removing appmenus for {!s}".format(vm),
                           file=sys.stderr)
             installed_appmenus = os.listdir(appmenus_dir)
-            directory_file = os.path.join(self.appmenus_dir(vm),
-                                          str(vm) + '-vm.directory')
-            installed_appmenus.remove(os.path.basename(directory_file))
-            if installed_appmenus:
-                appmenus_to_remove_fnames = [os.path.join(appmenus_dir, x)
-                                             for x in installed_appmenus]
-                try:
-                    desktop_menu_cmd = ['xdg-desktop-menu', 'uninstall']
-                    if not refresh_cache:
-                        desktop_menu_cmd.append('--noupdate')
-                    desktop_menu_cmd.append(directory_file)
-                    desktop_menu_cmd.extend(appmenus_to_remove_fnames)
-                    desktop_menu_env = os.environ.copy()
-                    desktop_menu_env['LC_COLLATE'] = 'C'
-                    subprocess.check_call(desktop_menu_cmd,
-                                          env=desktop_menu_env)
-                except subprocess.CalledProcessError:
-                    if hasattr(vm, 'log'):
-                        vm.log.warning(
-                            "Problem removing appmenus for %s", vm.name)
-
-                    else:
-                        print(
-                            "Problem removing appmenus for {!s}".format(vm),
-                            file=sys.stderr)
+            self._do_remove_appmenus(vm, installed_appmenus, appmenus_dir,
+                    refresh_cache)
             shutil.rmtree(appmenus_dir)
 
         if refresh_cache:
@@ -485,9 +592,8 @@ class Appmenus(object):
                     'qubes-start.desktop.template'))
 
         source_whitelist_filename = 'vm-' + AppmenusSubdirs.whitelist
-        if src and ('default-menu-items' in src.features or \
-                os.path.exists(os.path.join(
-                    basedir, src.name, source_whitelist_filename))):
+        if src and ('default-menu-items' in src.features or os.path.exists(
+                os.path.join(basedir, src.name, source_whitelist_filename))):
             vm.log.info("Creating default whitelisted apps list: {0}".
                         format(basedir + '/' + vm.name + '/' +
                                AppmenusSubdirs.whitelist))
@@ -629,9 +735,6 @@ parser.add_argument(
     '--i-understand-format-is-unstable', dest='fool',
     action='store_true',
     help='required pledge for --get-available')
-parser.add_argument(
-    'domains', metavar='VMNAME', nargs='+',
-    help='VMs on which perform requested actions')
 parser.add_argument(
     '--file-field', action='append', dest='fields',
     help='File field to append to output for --get-available; can be used'
@@ -640,8 +743,13 @@ parser.add_argument(
 parser.add_argument(
     '--template', action='store',
     help='Use the following template for listed domains instead of their '
-         'actual template. Requires --get-available.'
-)
+         'actual template. Requires --get-available.')
+parser.add_argument(
+    '--all', action='store_true', dest='all_domains',
+    help='perform the action on all qubes')
+parser.add_argument(
+    'domains', metavar='VMNAME', nargs='*', default=[],
+    help='VMs on which perform requested actions')
 
 
 def retrieve_list(path):
@@ -660,16 +768,30 @@ def retrieve_list(path):
 def main(args=None, app=None):
     """main function for qvm-appmenus tool"""
     args = parser.parse_args(args=args, app=app)
+    if not args.all_domains and not args.domains:
+        parser.error("one of the arguments --all VMNAME is required")
     appmenus = Appmenus()
     if args.source is not None:
         args.source = args.app.domains[args.source]
     if args.template is not None:
         args.template = args.app.domains[args.template]
-    for vm in args.domains:
+    if args.all_domains:
+        if args.remove:
+            domains = [vm for vm in os.listdir(os.path.abspath(basedir))
+                       if os.path.isdir(os.path.join(basedir, vm))]
+        else:
+            domains = args.app.domains
+    else:
+        domains = args.domains
+    for vm in domains:
+        if str(vm) == 'dom0':
+            continue
         # allow multiple actions
         # for remove still use just VM name (str), because VM may be already
-        # removed
+        # removed.
         if args.remove:
+            if isinstance(vm, qubesadmin.vm.QubesVM):
+                vm = vm.name
             appmenus.appmenus_remove(vm)
             appmenus.appicons_remove(vm)
             try:
@@ -678,7 +800,8 @@ def main(args=None, app=None):
                 pass
         # for other actions - get VM object
         if not args.remove:
-            vm = args.app.domains[vm]
+            if not isinstance(vm, qubesadmin.vm.QubesVM):
+                vm = args.app.domains[vm]
             if args.init:
                 appmenus.appmenus_init(vm, src=args.source)
             if args.get_whitelist:
