@@ -69,6 +69,7 @@ class TestVM(object):
     def __init__(self, name, klass, **kwargs):
         self.running = False
         self.is_template = False
+        self.template_for_dispvms=False
         self.name = name
         self.klass = klass
         self.log = logging.getLogger('qubesappmenus.tests')
@@ -126,6 +127,10 @@ class TC_00_Appmenus(unittest.TestCase):
             self.basedir)
         self.basedir_patch.start()
 
+    def _make_desktop_name(self, vm, appmenu_basename):
+        return os.path.join(self.ext.appmenus_dir(vm),
+                            self.ext.desktop_name(vm, appmenu_basename))
+
     def tearDown(self):
         self.basedir_patch.stop()
         self.basedir_obj.cleanup()
@@ -234,8 +239,7 @@ class TC_00_Appmenus(unittest.TestCase):
                 'test-data/evince.desktop.template'))
         self.ext.appmenus_create(appvm, refresh_cache=False)
         self.ext.appicons_create(appvm)
-        evince_path = os.path.join(
-            self.ext.appmenus_dir(appvm), appvm.name + '-evince.desktop')
+        evince_path = self._make_desktop_name(appvm, 'evince.desktop')
         self.assertPathExists(evince_path)
         with open(evince_path, 'rb') as f:
             self.assertEqual(
@@ -278,8 +282,7 @@ class TC_00_Appmenus(unittest.TestCase):
                 'test-data/evince.desktop.template').
                 replace(b'Document Viewer', b'Random Viewer'))
         self.ext.appmenus_update(appvm)
-        evince_path = os.path.join(
-            self.ext.appmenus_dir(appvm), appvm.name + '-evince.desktop')
+        evince_path = self._make_desktop_name(appvm, 'evince.desktop')
         self.assertPathExists(evince_path)
         with open(evince_path, 'rb') as f:
             self.assertEqual(
@@ -290,8 +293,7 @@ class TC_00_Appmenus(unittest.TestCase):
                 f.read()
             )
 
-        xterm_path = os.path.join(
-            self.ext.appmenus_dir(appvm), appvm.name + '-xterm.desktop')
+        xterm_path = self._make_desktop_name(appvm, 'xterm.desktop')
         self.assertPathExists(xterm_path)
         with open(xterm_path, 'rb') as f:
             self.assertEqual(
@@ -301,6 +303,62 @@ class TC_00_Appmenus(unittest.TestCase):
                 f.read()
             )
 
+    def test_007_created_dispvm(self):
+        tpl = TestVM('test-inst-tpl',
+            klass='TemplateVM',
+            virt_mode='pvh',
+            updateable=True,
+            provides_network=False,
+            label=self.app.labels[1])
+        self.ext.appmenus_init(tpl)
+        appvm = TestVM('test-inst-dvm',
+            klass='AppVM',
+            template=tpl,
+            virt_mode='pvh',
+            updateable=False,
+            provides_network=False,
+            template_for_dispvms=True,
+            label=self.app.labels[1])
+        appvm.features['appmenus-dispvm'] = '1'
+        self.ext.appmenus_init(appvm)
+        with open(os.path.join(self.ext.templates_dirs(tpl)[0],
+                'evince.desktop'), 'wb') as f:
+            f.write(pkg_resources.resource_string(__name__,
+                'test-data/evince.desktop.template'))
+        self.ext.appmenus_create(appvm, refresh_cache=False)
+        self.ext.appicons_create(appvm)
+        appmenus_dir = self.ext.appmenus_dir(appvm)
+
+        # "Template (disp)" menu
+        dirfile_path = os.path.join(appmenus_dir,
+            'qubes-vm-directory-test-inst-dvm.directory')
+        self.assertPathExists(dirfile_path)
+        with open(dirfile_path, 'rb') as f:
+            content = f.read()
+            self.assertIn(b'Name=Template (disp): test-inst-dvm\n', content)
+        evince_path = os.path.join(appmenus_dir,
+            'org.qubes-os.vm.test-inst-dvm.evince.desktop')
+        self.assertPathExists(evince_path)
+        with open(evince_path, 'rb') as f:
+            content = f.read()
+            self.assertNotIn(b'X-Qubes-NonDispvmExec=', content)
+            self.assertIn(b'X-Qubes-DispvmExec=', content)
+
+        # "Disposable" menu
+        dirfile_path = os.path.join(appmenus_dir,
+            'qubes-dispvm-directory-test-inst-dvm.directory')
+        self.assertPathExists(dirfile_path)
+        with open(dirfile_path, 'rb') as f:
+            content = f.read()
+            self.assertIn(b'Name=Disposable: test-inst-dvm\n', content)
+        evince_path = os.path.join(appmenus_dir,
+            'org.qubes-os.dispvm.test-inst-dvm.evince.desktop')
+        self.assertPathExists(evince_path)
+        with open(evince_path, 'rb') as f:
+            content = f.read()
+            self.assertIn(b'X-Qubes-NonDispvmExec=', content)
+            self.assertNotIn(b'X-Qubes-DispvmExec=', content)
+
     def test_100_get_appmenus(self):
         self.maxDiff = None
         def _run(service, **kwargs):
@@ -426,8 +484,9 @@ class TC_00_Appmenus(unittest.TestCase):
             provides_network=False,
             label=self.app.labels[1])
         self.ext.appmenus_init(tpl)
-        with open(os.path.join(self.ext.templates_dirs(tpl)[0],
-                'evince.desktop'), 'wb') as f:
+        old_path = os.path.join(self.ext.templates_dirs(tpl)[0],
+                                'evince.desktop')
+        with open(old_path, 'wb') as f:
             f.write(pkg_resources.resource_string(__name__,
                 'test-data/evince.desktop.template'))
         appvm = TestVM('test-inst-app',
@@ -437,10 +496,22 @@ class TC_00_Appmenus(unittest.TestCase):
             updateable=False,
             provides_network=False,
             label=self.app.labels[1])
+        appmenus_dir = self.ext.appmenus_dir(appvm)
+        self.assertEqual(appmenus_dir, self.basedir + '/test-inst-app/apps')
         self.ext.appmenus_init(appvm)
+        os.mkdir(appmenus_dir, mode=0o700)
+        self.assertEqual(os.listdir(appmenus_dir),[])
+        should_be_deleted = []
+        for i in 'desktop', 'directory':
+            for j in 'bad', 'org.qubes-os':
+                bad_path = os.path.join(appmenus_dir, j + '.' + i)
+                with open(bad_path, 'wb') as f:
+                    f.write(b'not a valid desktop file')
+                should_be_deleted.append(bad_path)
         self.ext.appmenus_create(appvm, refresh_cache=False)
-        evince_path = os.path.join(
-            self.ext.appmenus_dir(appvm), 'test-inst-app-evince.desktop')
+        for i in should_be_deleted:
+            self.assertPathNotExists(i)
+        evince_path = self._make_desktop_name(appvm, 'evince.desktop')
         self.assertPathExists(evince_path)
         with open(evince_path, 'rb') as f:
             self.assertEqual(
@@ -450,18 +521,30 @@ class TC_00_Appmenus(unittest.TestCase):
                 f.read()
             )
 
-        mock_subprocess.assert_called_once()
-        args = mock_subprocess.call_args[0][0]
+        retval = mock_subprocess.mock_calls
+        (_,(first_call,),_,), (_,(second_call,),_,), (_,(third_call,),_,), = retval
+        assert should_be_deleted[0].endswith('.desktop')
         self.assertEqual(
-            args[:3],
+            first_call[:],
+            ['xdg-desktop-menu', 'uninstall', '--noupdate', should_be_deleted[2], should_be_deleted[0]])
+        self.assertEqual(
+            second_call[:],
+            ['xdg-desktop-menu', 'uninstall', '--noupdate', should_be_deleted[3], should_be_deleted[1]])
+        self.assertEqual(
+            third_call[:3],
             ['xdg-desktop-menu', 'install', '--noupdate'])
-        prefix = self.basedir + '/test-inst-app/apps/test-inst-app-'
-        self.assertEqual(sorted(args[3:]), [
-            prefix + 'evince.desktop',
-            prefix + 'qubes-start.desktop',
-            prefix + 'qubes-vm-settings.desktop',
-            prefix + 'vm.directory'
-        ])
+        prefix = os.path.join(appmenus_dir, 'org.qubes-os.')
+        self.maxDiff = None
+        new_files = [
+            os.path.join(appmenus_dir, 'qubes-vm-directory-test-inst-app.directory'),
+            prefix + 'qubes-vm-settings.test-inst-app.desktop',
+            prefix + 'vm.test-inst-app.evince.desktop',
+            prefix + 'vm.test-inst-app.qubes-start.desktop',
+        ]
+        self.assertEqual(third_call[3], new_files[0])
+        self.assertEqual(set(third_call[4:]), set(new_files[1:]))
+        self.assertEqual(sorted(os.listdir(appmenus_dir)),
+                         sorted(os.path.basename(i) for i in new_files))
 
     @unittest.mock.patch('subprocess.check_call')
     def test_121_create_appvm_with_whitelist(self, mock_subprocess):
@@ -488,8 +571,7 @@ class TC_00_Appmenus(unittest.TestCase):
             label=self.app.labels[1])
         self.ext.appmenus_init(appvm)
         self.ext.appmenus_create(appvm, refresh_cache=False)
-        evince_path = os.path.join(
-            self.ext.appmenus_dir(appvm), 'test-inst-app-evince.desktop')
+        evince_path = self._make_desktop_name(appvm, 'evince.desktop')
         self.assertPathExists(evince_path)
         with open(evince_path, 'rb') as f:
             self.assertEqual(
@@ -504,11 +586,12 @@ class TC_00_Appmenus(unittest.TestCase):
         self.assertEqual(
             args[:3],
             ['xdg-desktop-menu', 'install', '--noupdate'])
-        prefix = self.basedir + '/test-inst-app/apps/test-inst-app-'
-        self.assertEqual(sorted(args[3:]), [
-            prefix + 'evince.desktop',
-            prefix + 'qubes-vm-settings.desktop',
-            prefix + 'vm.directory'
+        prefix = self.basedir + '/test-inst-app/apps/org.qubes-os.'
+        self.maxDiff = None
+        self.assertEqual(args[3:], [
+            self.basedir + '/test-inst-app/apps/qubes-vm-directory-test-inst-app.directory',
+            prefix + 'vm.test-inst-app.evince.desktop',
+            prefix + 'qubes-vm-settings.test-inst-app.desktop',
         ])
 
     def test_130_process_appmenus_templates(self):
@@ -569,3 +652,6 @@ class TC_00_Appmenus(unittest.TestCase):
 
 def list_tests():
     return (TC_00_Appmenus,)
+
+if __name__ == '__main__':
+    unittest.main()
